"""Tools for placing figures (not axes) on screen."""
import functools
import json
import platform
import warnings
from pathlib import Path

import matplotlib as mpl
from matplotlib import pyplot as plt

from mpl_tools import is_notebook_or_qt, is_using_interactive_backend

_FIG_GEOMETRIES_PATH = "./.fig_layout"


@functools.wraps(warnings.warn)
def warn(*args, **kwargs):
    """Avoid warnings printing the source code for itself.

    For example, we don't want `warnings.warn("Something went wrong")`
    to print that whole source line, which would duplicate the actual message.
    Using the `stacklevel` arg of `warn` doesn't cut it, because we don't
    necessarily want to climb one level up.

    Refs:
    - <https://stackoverflow.com/a/26433913>
    - <https://stackoverflow.com/a/2187390>
    """
    kwargs["stacklevel"] = kwargs.get("stacklevel", 1) + 1

    def frmt_warning(msg, category, filename, lineno, file=None, line=None):
        return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, msg)

    # Set custom frmt_warning (before calling warn), but restore original afterwards
    original = warnings.formatwarning
    warnings.formatwarning = frmt_warning
    warnings.warn(*args, **kwargs)
    warnings.formatwarning = original


def freshfig(num=None, place=True, **kwargs):
    """Create/clear figure, place it, call `plt.subplots(**kwargs)`.

    If `figure(num)` exists, it is **cleared** before calling `subplots`.
    Unlike closing/opening, this keeps its position and size.
    Note that `num` can also be a string, i.e. the label of the figure.

    If `place==2`, the figure placement is `load`ed, provided the figure's
    position has previously been `save`d. If `place==1`, the active placement
    only happens if the figure did not yet exist.
    If `figsize` is among the `kwargs` it will still be overruled
    by the figure placement, as it should (it is quite convenient when
    working with scripts that *also* run in Jupyter).

    .. note:
        Active placement (including re-sizing) on the `mpl` backend **MacOSX**
        [does not work](https://stackoverflow.com/a/30180994). However, sizing
        **upon creation** works, and Qt5Agg (or TkAgg) can also be installed.

    Example
    -------
    >>> freshfig(1, nrows=2, sharex=True)  # doctest: +NORMALIZE_WHITESPACE
    (<Figure size 640x480 with 2 Axes>,
    array([<AxesSubplot:>, <AxesSubplot:>], dtype=object))
    """
    already_open = plt.fignum_exists(num)

    fig = plt.figure(num=num, figsize=kwargs.pop("figsize", None))

    # Deal with warning bug
    # https://github.com/matplotlib/matplotlib/issues/9970
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        fig.clf()  # <=> fig.clear()

    if (
        place > 1 or (place and not already_open)
        and num is not None  # It makes little sense to load placement
                             # for the figure number resulting from None
    ):
        load(fignum=num)

    _, ax = plt.subplots(num=fig.number, **kwargs)
    return fig, ax


def _get_fig(fignum=None):
    """Get fig handle from number OR handle."""
    if fignum is None:
        return plt.gcf()
    elif isinstance(fignum, mpl.figure.Figure):
        return fignum
    else:
        return plt.figure(fignum)


class FigManagerDoesNotExist(Exception):
    """Signal that figure placement ain't possible with this mpl backend.

    Use exception (a custom subclass for finesse and thereby safety) to
    propagate error up stack to where we want to treat them. Consider: the
    alternative solution of returning some special value to signal an error
    would necessitate an additional `if (val==SPECIAL): return SPECIAL` clause
    at each intervening-level, even for functions that are not supposed to
    return anything.
    """


def _get_fmw(fignum):
    fig = _get_fig(fignum)
    try:
        fmw = fig.canvas.manager.window
        # fmw = plt.get_current_fig_manager().window
    except AttributeError:
        raise FigManagerDoesNotExist(
            "Cannot programmatically manipulate figure windows"
            f" with the current mpl. backend ({mpl.get_backend()})")
    return fmw


def _get_geo1(fignum):
    fmw = _get_fmw(fignum)
    try:
        # For Qt4Agg/Qt5Agg
        return dict(
            w=fmw.width(),
            h=fmw.height(),
            x=fmw.x(),
            y=fmw.y(),
        )
    except Exception:
        # For TkAgg
        return fmw.geometry()


def _set_geo1(fignum, geo):
    fmw = _get_fmw(fignum)
    try:
        # For Qt4Agg/Qt5Agg
        fmw.setGeometry(geo['x'], geo['y'], geo['w'], geo['h'])
    except Exception:
        # For TkAgg
        # geo = "{w:.0f}x{h:.0f}+{x:.0f}+{y:.0f}".format(**geo)
        fmw.geometry(newGeometry=geo)


def save(path=_FIG_GEOMETRIES_PATH, append_host=True):
    """Save current figure layout."""
    if append_host:
        path = ".".join([path, platform.node()])

    try:
        # Use labels if defined, else numbers
        lbls = [lbl or num for lbl, num in
                zip(plt.get_figlabels(), plt.get_fignums())]
        placements = {k: _get_geo1(k) for k in lbls}
    except FigManagerDoesNotExist as e:
        warn(str(e))
    else:
        # Get old placements
        if Path(path).is_file():
            with open(path, "r") as file:
                old = json.load(file)
        else:
            old = {}
        # Write old + new placements
        with open(path, "w") as file:
            file.write(json.dumps({**old, **placements}))


def load(path=_FIG_GEOMETRIES_PATH, append_host=True, fignum=None):
    """Load/set figure layout."""
    if is_notebook_or_qt or not is_using_interactive_backend():
        return  # quietly

    if append_host:
        path = ".".join([path, platform.node()])

    # Suggest saving layout
    if not Path(path).is_file():
        # TODO: deactivate until this is resolved:
        # https://stackoverflow.com/q/66388579
        # warn(f"Consider using {__name__}.save()"
        #      " for persistent figure layout.", stacklevel=2)
        return

    with open(path, "r") as file:
        placements = json.load(file)

    # Cast nums to int
    for key in list(placements):
        try:
            lbl = int(key)
        except ValueError:
            lbl = key
        placements[lbl] = placements.pop(key)

    try:
        for lbl in placements:
            if fignum is None or lbl == fignum:
                _set_geo1(lbl, placements[lbl])
    except FigManagerDoesNotExist as e:
        warn(str(e))


def show_figs(fignums=None):
    """Move all fig windows to top."""
    # Validate fignums
    if fignums is None:
        fignums = plt.get_fignums()
    try:
        fignums = list(fignums)
    except TypeError:
        fignums = [fignums]

    try:
        for f in fignums:
            fmw = _get_fmw(f)
            fmw.attributes('-topmost', 1)  # Bring to front, but
            fmw.attributes('-topmost', 0)  # don't keep in front
    except FigManagerDoesNotExist as e:
        warn(str(e))


def get_screen_size():
    """Get **available** screen size/resolution.

    NB: This might not always work that well,
    especially since the method used depends on the backend.

    Consider using non-mpl method: https://pypi.org/project/screeninfo
    """
    if mpl.get_backend().startswith('Qt'):
        try:
            # Ref spyder/widgets/shortcutssummary.py
            from qtpy.QtWidgets import QDesktopWidget  # type: ignore
            widget = QDesktopWidget()
            sc = widget.availableGeometry(widget.primaryScreen())
            x0, y0, w0, h0 = sc.x(), sc.y(), sc.width(), sc.height()

        except ImportError:
            # https://stackoverflow.com/a/29039755/38281
            fig = plt.figure()  # new, blank (cheap to resize) figure
            mgr = plt.get_current_fig_manager()
            # Now we full-screen the new figure, and pause for effect.
            # NB: both of these actions results in the figure being shown.
            mgr.full_screen_toggle(); plt.pause(0.2)  # noqa
            sc = mgr.canvas
            x0, y0, w0, h0 = sc.x(), sc.y(), sc.width(), sc.height()
            # mgr.full_screen_toggle()
            plt.close(fig)

    elif mpl.get_backend() == "TkAgg":
        # https://stackoverflow.com/a/42951711/38281
        mgr = plt.get_current_fig_manager()
        x0, y0, w0, h0 = (0, 0) + mgr.window.wm_maxsize()
        # w, h = mgr.window.winfo_screenwidth(), mgr.window.winfo_screenheight()
    else:
        print("Warning: could not detect screen size for this mpl backend.")
        print("Using a default size.")
        x0, y0, w0, h0 = 30, 30, 800, 600

    return x0, y0, w0, h0


def loc01(fignum=None, x=None, y=None, w=None, h=None):
    """Place figure on screen, in relative coordinates ∈ [0, 1].

    Only works on Linux?
    """
    try:
        fmw = _get_fmw(fignum)
    except FigManagerDoesNotExist as e:
        warn(str(e))
        return

    x0, y0, w0, h0 = get_screen_size()

    # It seems the window footers are not taken into account
    # by the geometry settings. Correct for this:
    footer = 0.028*(h0+y0)

    # Current values (Qt4Agg only!):
    w = w if w is not None else fmw.width() / w0
    h = h if h is not None else fmw.height()/h0
    x = x if x is not None else fmw.x()     / w0
    y = y if y is not None else fmw.y()     / h0

    x = x0 + x*w0
    y = y0 + y*h0 + footer
    w = w*w0
    h = h*h0 - footer

    try:  # For Qt4Agg/Qt5Agg
        fmw.setGeometry(x, y, w, h)
    except: # noqa # For TkAgg
        geo = f"{int(w)}x{int(h)}+{int(x)}+{int(y)}"
        fmw.geometry(newGeometry=geo)


def loc(loc, fignum=None):
    """Place figure on screen.

    - loc: string that defines the figures new geometry, given either as
       * NW, E, ...
       * 4 digits (as str or int) to define grid M,N,i,j.

    Example
    -------
    >>> N = 3
    >>> for i in 1 + np.arange(N):
    ...     xy = str(N)*2 + str(i)*2
    ...     loc(xy, i)
    """
    loc = str(loc)
    loc = loc.replace(",", "")
    if not loc[:4].isnumeric():
        if loc.startswith('NW'):
            loc = '2211'
        elif loc.startswith('SW'):
            loc = '2221'
        elif loc.startswith('NE'):
            loc = '2212'
        elif loc.startswith('SE'):
            loc = '2222'
        elif loc.startswith('W'):
            loc = '1211'
        elif loc.startswith('E'):
            loc = '1212'
        elif loc.startswith('S'):
            loc = '2121'
        elif loc.startswith('N'):
            loc = '2111'

    # Split digits
    M, N = int(loc[0]), int(loc[1])
    if loc[3] == '-':
        i1, i2 = int(loc[2]), int(loc[4])
    else:
        i1, i2 = int(loc[2]), int(loc[2])
    if loc[-2] == '-':
        j1, j2 = int(loc[-3]), int(loc[-1])
    else:
        j1, j2 = int(loc[-1]), int(loc[-1])
    # Validate
    assert M >= i2 >= i1 > 0, "The specified col index is invalid."
    assert N >= j2 >= j1 > 0, "The specified row index is invalid."

    # Place
    di = i2-i1+1
    dj = j2-j1+1
    loc01(fignum, (j1-1)/N, (i1-1)/M, dj/N, di/M)



"""Tools for placing axes in a figure."""
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.artist import allow_rasterization

# from mpl_tools.misc import *

__all__ = ["align_ax_with", "anchor_axes", "set_ax_size", "trans2fig"]


def set_ax_size(ax, w, h):
    """Width/Height in display (pixel) coords."""
    fig = ax.figure
    w, h = fig.transFigure.inverted().transform([w, h])
    x, y, _, _ = ax.get_position().bounds
    ax.set_position([x, y, w, h])


def align_ax_with(ax, bbox, loc, pad=4):
    """Set the location of ax relative to bbox.

    This is similar to legend placement, where bbox_to_anchor
    (here just bbox) defines the anchor position (as a box),
    and loc specifies the relative placement as the edge(s) to be aligned.

    However, here, loc should be one of NW, NE, SW, SE, E, W, N, S (as in Matlab),
    and each letter may appended by ``+`` (e.g. N+W+),
    meaning that the edge of ax is placed
    *outside* the opposite edge of (the container) bbox,
    with padding specified by pad.
    """
    # Get new bbox placed with mpl builtin tool
    B = ax.bbox.anchored(loc.replace("+", ""), container=bbox)
    # Adjust for + flags
    x, y, w, h = B.bounds
    if "W+" in loc:
        x += pad + bbox.width
    if "E+" in loc:
        x -= pad + bbox.width
    if "S+" in loc:
        y -= pad + h
    if "N+" in loc:
        y += pad + h
    B = mpl.transforms.Bbox.from_bounds(x, y, w, h)
    # Convert to figure coordinates
    # B = B.inverse_transformed(ax.figure.transFigure) # deprecated
    B = B.transformed(ax.figure.transFigure.inverted())
    # Set
    ax.set_position(B)


# TODO: make use of this in the above funcs
def trans2fig(axis, rect, from_data=True):
    """Transform (data or axis) coordinates to figure coordinates.

    Inspired by: https://stackoverflow.com/a/17478227/38281.

    Example
    -------
    >>> rect = trans2fig(ax, [x,y,w,h]) # doctest: +SKIP
    ... ax2 = ax.figure.add_axes(rect)
    """
    x, y, w, h = rect

    FT = axis.figure.transFigure.inverted().transform

    if from_data:
        # Transform: data-->display-->figure
        def T(xy):
            return FT(axis.transData.transform(xy))
    else:
        # Transform: axes-->display-->figure
        def T(xy):
            return FT(axis.transAxes.transform(xy))

    x, y = T((x, y))
    w, h = T([w, h]) - T([0, 0])  # affine transform wrt. 0

    return x, y, w, h


def anchor_axes(ax, get_anchor, loc="NW+"):
    """Attach axes to an anchoring bbox, and keep its size fixed.

    This is done by updating its placement whenever draw() is called,
    e.g. when the figure is resized.

    See also mpl_toolkits.axes_grid1.{Divider,Size}, demonstrated by:
    https://matplotlib.org/gallery/axes_grid1/demo_fixed_size_axes.html
    """
    # Save size
    size = ax.bbox.width, ax.bbox.height

    # Patch the Axes instance's draw() method.
    @allow_rasterization
    def draw(self, renderer):
        set_ax_size(self, *size)
        align_ax_with(self, get_anchor(), loc)
        _draw(renderer)
    _draw = ax.draw
    ax.draw = draw.__get__(ax)


def get_legend_bbox(ax):
    """Get legend's bbox in pixel ("display") coords."""
    # Must pause/draw before bbox can be known
    def inner():
        plt.draw()
        leg = ax.get_legend()
        bbox = leg.get_window_extent()
        # bbox = leg.get_frame().get_bbox()
        return bbox
    return inner



"""Documentation for mpl-tools.

## Dev guide
Also see `dev_guide`

## API reference
The rendered docstrings can be browsed
through the following links, which are also available in the left sidebar.
"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore

# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
__version__ = importlib_metadata.version(__name__)


import matplotlib as mpl

try:
    __IPYTHON__  # type: ignore
    from IPython import get_ipython  # type: ignore
    ip = str(type(get_ipython())).lower()
    is_notebook_or_qt = (
        ("zmq" in ip)  # local Jupyter, binder, Kaggle
        or ("colab" in ip)  # Google Colab
    )
except (NameError, ImportError):
    is_notebook_or_qt = False


def is_using_interactive_backend():
    """Check if `mpl` is (currently) using an interactive backend.

    This means that the figures may be shown on screen.
    But only if you've done `plt.ion` so that `mpl.is_interactive`,
    or after calling `plt.show`.

    Note on jupyter notebook backends:
    - `%matplotlib notebook` (`nbAgg`) is interactive.
    - `%matplotlib inline` is not, but still shows up in the notebook.
      It can also be updated; see `dapper/tools/viz:plot_pause()`.
      However, it does not allow for mouse interaction,
      as it is merely a static picture.
      It also uses different dpi than `nbAgg`, ref
      https://github.com/matplotlib/matplotlib/issues/4853
    - `%matplotlib widgets` is for Jupyterlab?
    """
    return mpl.get_backend() in mpl.rcsetup.interactive_bk



"""Science-related mpl tools."""

import matplotlib.colors
import matplotlib.ticker
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Ellipse
from numpy.linalg import eigh

from mpl_tools.fig_layout import freshfig


def cov_ellipse(ax, mu, sigma, **kwargs):
    r"""Draw ellipse corresponding to (Gaussian) 1-sigma countour of cov matrix.

    [Inspiration](https://stackoverflow.com/q/17952171)

    Example
    -------
    >>> fig, ax = plt.subplots()
    >>> x, y, C = 10, 20, [[1, .3], [.3, 1]]
    >>> _ = ax.set(xlim=(x-1, x+1), ylim=(y-1, y+1))
    >>> ellipse = cov_ellipse(ax, (x, y), C,
    ...                       fc='none', ec='r', lw=4, label='$1\\sigma$')
    """
    # Cov --> Width, Height, Theta
    vals, vecs = eigh(sigma)
    x, y = vecs[:, -1]  # x-y components of largest (last) eigenvector
    theta = np.degrees(np.arctan2(y, x))
    theta = theta % 180

    h, w = 2 * np.sqrt(vals.clip(0))

    # Get artist
    e = Ellipse(mu, w, h, theta, **kwargs)

    ax.add_patch(e)
    e.set_clip_box(ax.bbox)  # why is this necessary?

    # Return artist
    return e


def axes_with_marginals(n_joint, n_marg, **kwargs):
    """Create a joint axis along with two marginal axes.

    Example
    -------
    >>> ax0, a_x, a_y = axes_with_marginals(4, 1)
    >>> x, y = np.random.randn(2,500)
    >>> _ = ax0.scatter(x,y)
    >>> _ = a_x.hist(x)
    >>> _ = a_y.hist(y,orientation="horizontal")
    """
    N = n_joint + n_marg

    # Method 1
    # fig, ((ax0, a_y), (a_x, _)) = plt.subplots(2,2,num=plt.gcf().number,
    # sharex='col',sharey='row',gridspec_kw={
    #     'height_ratios':[n_joint,n_marg],
    #     'width_ratios' :[n_joint,n_marg]})
    # _.set_visible(False) # Actually removing would bug the axis ticks etc.

    # Method 2
    gs   = GridSpec(N, N, **kwargs)
    fig  = plt.gcf()
    ax0 = fig.add_subplot(gs[n_marg:N, 0:n_joint])
    a_x = fig.add_subplot(gs[0:n_marg, 0:n_joint], sharex=ax0)
    a_y = fig.add_subplot(gs[n_marg:N, n_joint:N], sharey=ax0)
    # Cannot delete ticks coz axis are shared
    plt.setp(a_x.get_xticklabels(), visible=False)
    plt.setp(a_y.get_yticklabels(), visible=False)

    return ax0, a_x, a_y


def matshow_discrete(X, fig_ax=None, cmap=None, mode="set", ndigits=8):
    """Do matshow, add **discrete colorbar**.

    Inspired by https://stackoverflow.com/a/60870122

    Example:
    >>> from scipy import sparse
    >>> D = sparse.diags([1, -2, 1], [-1, 0, 1], shape=(9, 9))
    >>> image, colorbar = matshow_discrete(D.A)
    """
    if isinstance(fig_ax, str):
        fig, ax = freshfig(fig_ax)
    elif fig_ax is None:
        fig, ax = freshfig("matshow_discrete")
    else:
        fig, ax = fig_ax

    X = np.asarray(X)
    m, M = X.min(), X.max()

    # Default cmap
    if cmap is None:
        if M == 0:
            cmap = "cool"
        elif m == 0:
            cmap = "autumn"
        elif m < 0 and M > 0:
            cmap = "coolwarm"
        else:
            cmap = "jet"

    if mode == "set":
        # Get unique, sorted list of values
        S = np.round(X, ndigits)
        S = set(S.flat)  # unique
        S = np.array(sorted(S))
        # Get boundaries below and above each value of S
        # (the exact amount of these margins don't matter).
        bins = (S[1:] + S[:-1])/2
        bins = np.concatenate(([m-1], bins, [M+1]))
        # Center ticks (vs bins)
        ticks = bins[:-1] + np.diff(bins)/2
        # Custom tick labels
        formatter = matplotlib.ticker.FuncFormatter(
            lambda x, idx: "None" if idx is None else "%.2f" % S[idx])  # type: ignore
        # Create cmap and norm
        if cmap == "coolwarm":
            # Ensure 0 corresponds to .5
            neg = np.linspace(0, .5, np.sum(S < 0) + 1)[:-1]
            pos = np.linspace(.5, 1, np.sum(S > 0) + 1)[1:]
            xx = np.concatenate([neg, [.5], pos])
        else:
            xx = np.linspace(0, 1, len(S))
        cmap = plt.get_cmap(cmap)(xx)
        cmap = matplotlib.colors.ListedColormap(cmap)
        norm = matplotlib.colors.BoundaryNorm(bins, cmap.N)

        im = ax.matshow(X, cmap=cmap, norm=norm)
        cb = fig.colorbar(im, ticks=ticks, format=formatter)

    elif mode == "linear":
        nlevels = 11
        ticks = np.linspace(m, M, nlevels)
        cmap = plt.get_cmap(cmap, nlevels)
        im = ax.matshow(X, cmap=cmap,
                        # Center outer ticks by stretching map
                        vmin=1.5*ticks[0] - .5*ticks[1],
                        vmax=1.5*ticks[-1] - .5*ticks[-2])
        cb = fig.colorbar(im, ticks=ticks)

    else:
        raise ValueError("Invalid mode.")

    return im, cb


def matshow_banded(bands, fig_ax=None, lower=True):
    """Plot matrix with `bands` as in `solveh_banded`, using `matshow_discrete`.

    Example:
    >>> bands = np.zeros((2, 10-2))
    >>> bands[0] = 2
    >>> bands[1, :-1] = -1
    >>> image, colorbar = matshow_banded(bands)
    """
    if not lower:
        raise NotImplementedError

    m = bands.shape[-1]
    D = np.zeros((m, )*2)
    for i, band in enumerate(bands):
        c = 1 if i else .5
        D += c * np.diag(band[:m-i], k=i)
    D += D.T

    return matshow_discrete(D, fig_ax)



"""Configures pytest (beyond the ini file)."""
import numpy
import pytest


@pytest.fixture(autouse=True)
def add_np(doctest_namespace):
    """Add numpy as np for doctests."""
    doctest_namespace["np"] = numpy



"""Toggle log-scale on and off. Provide button for it."""
import warnings

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.widgets import CheckButtons  # Button

from mpl_tools.ax_placement import anchor_axes, get_legend_bbox
from mpl_tools.misc import thousands


def add_log_toggler(ax, pos="leg:S+W", ylim=None):
    """Add button that toggles log. scale."""
    fig = ax.figure

    if mpl.get_backend() == "Qt5Agg":
        print("Warning: Qt5Agg is used, which is not very fast with this function")

    # button_ax: size and creation
    size = _xFontsize(mpl.rcParams['font.size'], fig, 10, 2.5)
    size = fig.transFigure.inverted().transform(size)
    rect = [.5, .5, *size]
    button_ax = fig.add_axes(rect, frameon=False, xticks=[], yticks=[])
    # button_ax: position
    if "leg:" in pos:
        pos = pos.replace("leg:", "")
        # Continuous re-positioning:
        anchor_axes(button_ax, get_legend_bbox(ax), pos)
        # Position once only (will scale with figure):
        # align_ax_with(button_ax, get_legend_bbox(ax)(), "NW+")
    else:
        anchor_axes(button_ax, lambda: ax.bbox, "NE")

    # Create button/checkmarks
    ax.log_toggler = CheckButtons(button_ax, ["Log scale"], [False])
    # Adjust checkmark style
    dh = .3
    for box, cross in zip(ax.log_toggler.rectangles, ax.log_toggler.lines):
        box.set_y(dh)
        box.set_height(1 - 2 * dh)
        cross[0].set_ydata([dh, 1 - dh])
        cross[1].set_ydata([dh, 1 - dh][::-1])

    # Set callback
    def toggler(_):
        return toggle_scale(ax, ylim, _toggle_button=False)
    ax.log_toggler.on_clicked(toggler)


def toggle_scale(ax, ylim=None, formatter=thousands, _toggle_button=True):
    """Toggle log. scale.

    _toggle_button is used internally. Do not set.
    """
    # Get status
    log_is_on = getattr(ax, "_log_is_on", False)

    # Toggle button
    # NB: Setting the active status of the button
    # will call this same function,
    # but with _toggle_button=False.
    if _toggle_button and hasattr(ax, "log_toggler"):
        ax.log_toggler.set_active(0)

    else:
        # Toggle scale
        ax._log_is_on = not log_is_on
        if log_is_on:
            ax.set_yscale("linear")
            ax.set_ylim(bottom=0)
            ax.yaxis.set_major_formatter(formatter)
        else:
            ax.set_yscale("linear")
            ax.autoscale(True, axis="y")
            if ylim:
                ax.set_ylim(ylim)
            ax.set_yscale("log")
            ax.yaxis.set_major_formatter(formatter)
        plt.draw()


def _xFontsize(fontsize, fig, *args):
    """Multiply by fontsize, in pixels (rather than points)."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        plt.pause(.1)
    fontsize = fig.canvas.renderer.points_to_pixels(fontsize)
    return tuple(a * fontsize for a in args)



"""Misc mpl tools."""
import inspect
import warnings
from math import sqrt
from pathlib import Path

import matplotlib as mpl
from matplotlib import pyplot as plt

thousands = mpl.ticker.StrMethodFormatter('{x:,.7g}')  # type: ignore


def axprops(dct):
    """Filter `dct` for properties associated with a plot axes.

    Example
    -------
    >>> def myplotter(ax, x, y, **kwargs):
    ...     ax.set(**axprops(kwargs))  # Pop the kwargs belonging to ax
    ...     ax.plot(x, y, kwargs)      # Use rest as line props.
    """
    # List of included axis properties
    props = ["title", "facecolor", "aspect"]
    # Append xyz-specific props
    for ax in ["x", "y", "z"]:
        for p in ["label", "ticks", "scale", "lim"]:
            props.append(ax + p)

    # intersection(dct,props)
    props = {p: dct.pop(p) for p in props if p in dct}

    return props


def fig_colorbar(fig, collections, *args, **kwargs):
    """Add colorbar to the right on a figure.

    Example
    -------
    >>> fig, (ax1, ax2) = plt.subplots(nrows=2)
    >>> cs = ax1.contourf(np.arange(24).reshape((3,-1)), cmap=plt.cm.PuBu_r)
    >>> # fig.colorbar(cs)          # => colorbar to the right of ax2
    >>> cb = fig_colorbar(fig, cs)  # => colorbar on the right of figure.
    """
    fig.subplots_adjust(right=0.8)
    cax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    cbar = fig.colorbar(collections, cax, *args, **kwargs)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        plt.pause(.1)
    return cbar


def nRowCol(nTotal, figsize=None, axsize=None):
    """Compute `(nrows, ncols)` such that `nTotal ≈ nrows*ncols`.

    Takes into account the `figsize` and `axsize`, either of which
    may be given as a tuple, or a number (the ratio width/height).
    Defaults are taken from `mpl.rcParams`.

    Examples
    --------
    >>> nRowCol(4)
    {'nrows': 2, 'ncols': 2}
    >>> nRowCol(4, (4, 1), (1, 1))
    {'nrows': 1, 'ncols': 4}
    >>> nRowCol(4, (4, 1), (4, 1))
    {'nrows': 2, 'ncols': 2}
    >>> nRowCol(12, (4, 3), (1, 1))
    {'nrows': 3, 'ncols': 4}
    """
    # Defaults
    if figsize is None:
        figsize = mpl.rcParams["figure.figsize"]

    if axsize is None:
        rc = {k: mpl.rcParams["figure.subplot." + k]
              for k in ["bottom", "left", "right", "top"]}
        axsize = (rc["right"] - rc["left"],
                  rc["top"] - rc["bottom"])

    # Compute ratios
    try:
        w, h = figsize
        fig_ratio = w / h
    except TypeError:
        fig_ratio = figsize
    try:
        w, h = axsize
        ax_ratio = w / h
    except TypeError:
        ax_ratio = axsize
    ratio = fig_ratio / ax_ratio

    # Main logic
    nrows = round(sqrt(nTotal / ratio)) or 1
    ncols = nTotal // nrows
    # ncols rounds down, so we might have to increase it.
    if nrows * ncols < nTotal:
        ncols += 1

    return {"nrows": nrows, "ncols": ncols}


# TODO: re-include NicePrint from struct-tools?
class FigSaver():
    """Simplify exporting a figure, especially when it's part of a series."""

    def __init__(self, script=None, basename=None, n=-1, ext='.pdf'):
        """Init.

        Parameters
        ----------
        script: str
            Name of dir to put figs in. Defaults to stem of `__file__` of caller.

        basename: str
            Name of figure files.

        n: int
           Starting index for file names.

        Example
        -------
        >>> fs = FigSaver(n=1, ext=".eps") # doctest: +SKIP
        ... ax.plot(xdata, ydata) # Add line to plot
        ... fs.save()             # Save
        ... ax.plot(xdata, ydata) # Add another line
        ... fs.save()
        """
        # Defaults
        if script is None:  # Get __file__ of caller
            script = inspect.getfile(inspect.stack()[1][0])
        if basename is None:
            basename = 'figure'
        # Prep save dir
        sdir = Path(script).stem
        Path(sdir).mkdir(parents=True, exist_ok=True)
        # Set state
        self.fname = sdir + basename
        self.n     = n
        self.ext   = ext

    @property
    def fullname(self):
        """Get full name of figure to be saved."""
        f = self.fname            # Abbrev
        if self.n >= 0:           # If indexing:
            f += '_n%d' % self.n  # Add index
        f += self.ext             # Add extension
        return f

    def save(self):
        """Save current figure. Increments the index."""
        f = self.fullname           # Abbrev
        print("Saving fig to:", f)  # Print
        plt.savefig(f)              # Save
        if self.n >= 0:             # If indexing:
            self.n += 1                 # Increment
            plt.pause(0.1)              # For safety



"""Toggle visisbility of plot elements on/off. Provide checkmarks for it."""
import itertools
import textwrap
import warnings

import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import transforms as mtransforms
from matplotlib.widgets import CheckButtons


def toggle_lines(ax=None, autoscl=True, numbering=False,
                 txtwidth=15, txtsize=None, state=None):
    """Make checkbuttons to toggle visibility of each line in current plot.

    - `autoscl`  : Rescale axis limits as required by currently visible lines.
    - `numbering`: Add numbering to labels.
    - `txtwidth` : Wrap labels to this length.

    The state of checkboxes can be inquired by
    >>> OnOff = [lh.get_visible() for lh in # doctest: +SKIP
    ...          ax.findobj(lambda x: isinstance(x,mpl.lines.Line2D))[::2]]
    """
    if ax is None:
        ax = plt.gca()
    if txtsize is None:
        txtsize = mpl.rcParams['font.size']

    # Get lines and their properties
    lines = {'handle': list(ax.get_lines())}
    for prop in ['label', 'color', 'visible']:
        lines[prop] = [plt.getp(x, prop) for x in lines['handle']]

    # Rm those that start with _
    not_ = [not x.startswith('_') for x in lines['label']]
    for prop in lines:
        lines[prop] = list(itertools.compress(lines[prop], not_))
    N = len(lines['handle'])

    # Adjust labels
    if numbering:
        lines['label'] = [str(i) + ': ' + x for i, x in enumerate(lines['label'])]
    if txtwidth:
        lines['label'] = [textwrap.fill(x, width=txtwidth) for x in lines['label']]

    # Set state. BUGGY? sometimes causes MPL complaints after clicking boxes
    if state is not None:
        state = np.array(state).astype(bool)
        lines['visible'] = state
        for i, x in enumerate(state):
            lines['handle'][i].set_visible(x)

    # Setup buttons
    # When there's many, the box-sizing is awful, but difficult to fix.
    W = 0.23 * txtwidth / 15 * txtsize / 10
    nBreaks = sum(x.count('\n') for x in lines['label'])  # count linebreaks
    H = min(1, 0.05 * (N + nBreaks))
    plt.subplots_adjust(left=W + 0.12, right=0.97)
    rax = plt.axes([0.05, 0.5 - H / 2, W, H])
    check = CheckButtons(rax, lines['label'], lines['visible'])

    # Adjust button style
    for i in range(N):
        check.rectangles[i].set(lw=0, facecolor=lines['color'][i])
        check.labels[i].set(color=lines['color'][i])
        if txtsize:
            check.labels[i].set(size=txtsize)

    # Callback
    def toggle_visible(label):
        ind    = lines['label'].index(label)
        handle = lines['handle'][ind]
        vs     = not lines['visible'][ind]
        handle.set_visible(vs)
        lines['visible'][ind] = vs
        if autoscl:
            _autoscale_based_on(ax, list(itertools.compress(
                lines['handle'], lines['visible'])))
        plt.draw()
    check.on_clicked(toggle_visible)

    # Return focus
    plt.sca(ax)

    # Must return (and be received) so as not to expire.
    return check


# https://stackoverflow.com/a/7396313
def _autoscale_based_on(ax, line_handles):
    """Autoscale axis based (only) on `line_handles`."""
    ax.dataLim = mtransforms.Bbox.unit()
    for iL, lh in enumerate(line_handles):
        xy = np.vstack(lh.get_data()).T
        ax.dataLim.update_from_data_xy(xy, ignore=(iL == 0))
    ax.autoscale_view()


def toggle_viz(*handles, prompt=False, legend=False, pause=0.0):
    """Toggle visibility of the graphics with handle `handles`."""
    are_viz = []
    for h in handles:

        # Core functionality: turn on/off
        is_viz = not h.get_visible()
        h.set_visible(is_viz)
        are_viz += [is_viz]

        # Legend updating. Basic version: works by
        #  - setting line's label (to /'_nolegend_' if off)
        #  - re-calling legend()
        if legend:
            if is_viz:
                try:
                    h.set_label(h.actual_label)
                except AttributeError:
                    pass
            else:
                h.actual_label = h.get_label()
                h.set_label('_nolegend_')
            # Legend refresh
            ax = h.axes
            with warnings.catch_warnings():
                warnings.simplefilter("error", category=UserWarning)
                try:
                    ax.legend()
                except UserWarning:
                    # If all labels are '_nolabel_' then ax.legend() throws warning,
                    # and quits before refreshing.
                    # => Refresh by creating/rm another legend.
                    ax.legend('TMP').remove()

    # Pause at where used (typically sequentially in script)
    if prompt:
        input("Press <Enter> to continue...")
    if pause > 0:
        plt.pause(pause)

    return are_viz



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mpl_tools']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.0', 'scipy']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

setup_kwargs = {
    'name': 'mpl-tools',
    'version': '0.2.36',
    'description': 'Tools for working with matplotlib',
    'long_description': '# mpl-tools\n\n[![Travis](https://travis-ci.org/patricknraanes/mpl-tools.svg?branch=master)](https://travis-ci.org/patricknraanes/mpl-tools)\n[![Coveralls](https://coveralls.io/repos/github/patricknraanes/mpl-tools/badge.svg?branch=master)](https://coveralls.io/github/patricknraanes/mpl-tools?branch=master)\n[![Hits.dwyl](http://hits.dwyl.com/patricknraanes/mpl-tools.svg)](http://hits.dwyl.com/patricknraanes/mpl-tools)\n[![PyPI](https://badge.fury.io/py/mpl-tools.svg)](https://badge.fury.io/py/mpl-tools)\n[![PyPI - Downloads](https://img.shields.io/pypi/dw/mpl-tools)](https://pypi.org/project/mpl-tools/0.1.5/)\n\nThis package provides some tools to work with [Matplotlib](https://matplotlib.org/).\n\n## Installation\n\n#### Normal installation\n\n```sh\npip install mpl-tools\n```\n\n#### For development\n\n**On Linux**:\n`git clone <this repo> ; make install`\n\n**On Windows (or Linux)**:\nDownload & extract this repo.\nInstall [poetry](https://python-poetry.org/docs/#installation).\nRun `poetry install`.\n\n**Note**: The above will install *poetry*,\nwhich then installs the source and its dependencies\ninto an isolated *virtual environment*.\nTo add packages, you should *not* use `pip install`,\nbut *poetry*, e.g. `poetry add --dev somepackage`.\n\n<!-- markdownlint-disable-file header-increment -->\n',
    'author': 'patricknraanes',
    'author_email': 'patrick.n.raanes@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/patricknraanes/mpl-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
